/************************************************************************
  UNIX işletim sisteminde C ile SEMAFOR ve ORTAK BELLEK ALANI KULLANIMI

                     Beycan KAHRAMAN         040020337

                     14.04.2006              Yrd. Doç. Dr. A. Şima UYAR
************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#define KEYSHM        10000        // Ortak Paylaşımlı Bellek Alanı
#define SEMKEY        8            // Semafor
#define COCUK_SAYISI  5
#define SAYI          20           // Çocuk Başına Düşen Dizi Elemanı
#define MAX_RANDOM    1000         // Dizinin İçeriği

/*-----------------------------------------------------------------
 Programın hazırlanmasında uygulamada verilen SEMAFOR kullanımı
 örneğinden büyük oranda yararlanılırken; ORTAK BELLEK
 ALANI KULLANIMI uygulamasından :
       1. ORTAK PAYLAŞIMLI DİZİ OLUŞTURMA
       2. DİZİYİ SİSTEME İADE ETME
 ifadeleri ilave edilmiştir.

 Ayrıca programın hazırlanmasında MergeSort Sıralama
 Algoritmasından da Yararlanılmıştır.
       1. Çocuklar Kendi Alanlarını Sıralarken MergeSort()
       2. Anne'de Tüm Dizi Birleştirilirken Merge()
 Fonksiyonları Kullanılmıştır.
-----------------------------------------------------------------*/

void mergeSort ( int, int, int * );
void merge ( int, int, int, int * );

void signal12 ( void )
{}

void sem_signal ( int semid, int val )
{
    struct sembuf semafor;

    semafor.sem_num=0;
    semafor.sem_op=val;
    semafor.sem_flg=1;
    semop(semid, &semafor,1);
}

void sem_wait ( int semid, int val )
{
    struct sembuf semafor;

    semafor.sem_num=0;
    semafor.sem_op=(-1*val);
    semafor.sem_flg=1;
    semop(semid, &semafor,1);
}

int main (void)
{
    int f = 1, i;
    int cocuk[COCUK_SAYISI];
    int sonsem = 0, pba = 0, *dizi = NULL;

    signal ( 12, (void *) signal12 );
    srand ( time(NULL) );

    for ( i=0; i<COCUK_SAYISI; ++i )
    {
        if ( f > 0 )
            f = fork ();
        if ( f == -1 )
        {
            printf("fork hatası...\n");
            exit ( 1 );
        }
        if ( f == 0 )
            break;
        else
            cocuk[i]=f;
    }

    if ( f > 0 ) /* anne */
    {
        sonsem = semget ( SEMKEY, 1, 0700|IPC_CREAT );
        semctl ( sonsem, 0, SETVAL, 0 );

// DİZİ OLUŞTURMA
        printf("    Anne     : 'Diziyi Oluşturuyorum.'\n", i+1);
        pba = shmget ( KEYSHM, sizeof(int)*COCUK_SAYISI*SAYI, 0700|IPC_CREAT );
        dizi = (int *) shmat ( pba, 0, 0 );

        for ( i=0; i<COCUK_SAYISI*SAYI; ++i )
            dizi[i] = rand()%MAX_RANDOM;
// ----------------

        sleep(1);

        printf("    Anne     : 'Çocukları Oluşturuyorum.'\n", i+1);
        for ( i=0; i<COCUK_SAYISI; ++i )
            kill(cocuk[i], 12);

        sem_wait(sonsem,COCUK_SAYISI);
        printf("    Anne     : 'Çocuklar Öldü.'\n", i+1);

// EKRANA YAZMA 1 ( Çocuklardan gelen hali )
        printf("    Anne     : 'Çocuklardan Dönen Hali:'\n", i+1);
        for ( i=0; i<COCUK_SAYISI*SAYI; ++i ){
            printf ( "%d ", dizi[i] );
            if ( i%SAYI == SAYI-1 )
                printf ( "\n" );
        }
// ----------------

// DİZİYİ BİRLEŞTİRME
        printf("    Anne     : 'Birleştiriyorum.'\n", i+1);
        for ( i=1; i<COCUK_SAYISI; ++i )
            merge ( 0, i*SAYI-1, (i+1)*SAYI-1, dizi );
// ----------------

// EKRANA YAZMA II ( Sıralanmış Hali )
        printf("    Anne     : 'Sıralanmış Hali:'\n", i+1);
        for ( i=0; i<COCUK_SAYISI*SAYI; ++i ){
            printf ( "%d ", dizi[i] );
            if ( i%SAYI == SAYI-1 )
                printf ( "\n" );
        }
// ----------------

        printf("    Anne     : 'Çıkıyorum.'\n", i+1);
        shmctl(pba,IPC_RMID,0);
        semctl(sonsem,0,IPC_RMID,0);
        exit ( 0 );
    }
    else /* cocuk */
    {
        pause();
        sonsem = semget(SEMKEY, 1,0);
        pba = shmget ( KEYSHM, sizeof(int)*COCUK_SAYISI*SAYI,0 );
        dizi = (int *) shmat ( pba, 0, 0 );

// i. BÖLÜMÜ SIRALAMA
        printf("    %d. Çocuk : 'Sıralıyorum.'\n", i+1 );
        mergeSort ( SAYI*i, SAYI*(i+1)-1, dizi );
// ----------------

        sem_signal(sonsem,1);
        exit ( 0 );
    }
    return(0);
}

void merge ( int bas, int orta, int son, int *dizi )
{
    int i = bas, j = orta + 1;
    int k, boyut = son - bas + 1;

    int *yeniDizi = (int *) malloc( boyut*sizeof(int) );

    for ( k = 0; k < boyut; ++k )
        if ( dizi[i] < dizi[j] && i <= orta || j > son )
            yeniDizi[k] = dizi[i++];
        else
            yeniDizi[k] = dizi[j++];

    for ( i = 0; i < boyut; ++i )
        dizi[bas + i] = yeniDizi[i];

    free ( yeniDizi );
}

void mergeSort ( int bas, int son, int *dizi )
{
    if ( bas == son ) return;

    int  orta = (bas + son) / 2;

    mergeSort ( bas, orta, dizi );
    mergeSort ( orta + 1, son, dizi );

    merge ( bas, orta, son, dizi );
}

